/*
 *    SimpleGraphicalClient.java
 *    Copyright (C) 2001 New Zealand Digital Library Project
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package org.nzdl.gsdl.SimpleGraphicalClient;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.io.IOException;
import java.util.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.text.*;
import javax.swing.text.html.*; 


// local libraries
import org.nzdl.gsdl.service.NzdlCollectionInfo; 
import org.nzdl.gsdl.service.NzdlQuery;
import org.nzdl.gsdl.service.NzdlRequest;
import org.nzdl.gsdl.service.NzdlResponse;
import org.nzdl.gsdl.service.NzdlResultSet;
import org.nzdl.gsdl.service.NzdlService;
import org.nzdl.gsdl.service.NzdlServiceClient;
import org.nzdl.gsdl.util.*;


/**
 * A Class representing the Panel in which the Querying action happens. 
 *
 * Does most of the actual `work' in the package.
 * 
 * @author Dave Nichols (daven@cs.waikato.ac.nz)
 * @author stuart yeates (say1@cs.waikato.ac.nz)
 * @version $Revision: 2281 $
 * @see org.nzdl.gsdl.service.SimpleGraphicalClient.SimpleGraphicalClient
 * @see org.nzdl.gsdl.service.SimpleGraphicalClient.CSModel
 * @see javax.swing.JPanel
 */

public class SearchPanel extends JPanel implements ActionListener, Constants
{

  CSModel csModel;
  NzdlService nzdl;

  JLabel collectionLabel;
  JButton collectionInfoButton;  
  JTextField searchTextField;
  JButton searchButton;
  JComboBox collectionList;  
  JPanel queryFormulationPanel, resultsPanel, dataPanel, optionsPanel; 
    JPanel queryTypePanel, searchControlPanel;
  JPanel collectionListPanel, searchTextFieldPanel, searchButtonPanel;
    JPopupMenu resultPopup;
    Result selectedResult;
    ResultModel resultListModel;
  JList   resultsList;   
    JCheckBox stemCheckBox, caseFoldCheckBox;
    JRadioButton rankedRadioButton, booleanRadioButton;
    ButtonGroup buttonGroup;
  JScrollPane scrollResultsPane;
  JScrollPane scrollDataPane;
  JFrame windowParent;
    //JEditorPane documentPane;
    JTextPane documentPane; // make this a JTextComponent ??
    JTextArea documentArea;
    JTextComponent documentComponent;
    HTMLEditorKit htmlEditorKit = new HTMLEditorKit();
    StyledEditorKit styledEditorKit = new StyledEditorKit();
    GMLDocument htmlDoc = new GMLDocument();
    DefaultStyledDocument defaultStyledDoc = new DefaultStyledDocument();
    final static GMLDocument blankDoc = new GMLDocument();
  
    /**
   * Construct and initialise a new SearchPanel
   */

  public SearchPanel( CSModel newCsModel, JFrame parent) 
  {
    super();
    windowParent = parent;
    csModel = newCsModel;
    nzdl = csModel.getNzdlService();
    setLayout( new BoxLayout(this, BoxLayout.Y_AXIS));
    setBorder(BorderFactory.createTitledBorder("Search"));
	
	
    searchTextField = new JTextField("Enter search terms here", 35) { 
	public boolean isRequestFocusEnabled(){ 
	  return true; 
	}
      };
    searchTextField.setText("Enter search terms here");
    searchTextField.setFont(searchTextFieldFont);
    searchTextField.setColumns(35);
    searchTextField.setBorder(BorderFactory.createEmptyBorder(5,5,5,5));
    searchButton = new JButton("Search");
	
    searchButton.addActionListener(this);
    //searchButton.setBorder(BorderFactory.createEmptyBorder(2,2,2,2));
    searchButton.setPreferredSize(new Dimension(80, 20));
	
    collectionListPanel = new JPanel();
    collectionListPanel.setBorder(BorderFactory.createEmptyBorder(5,5,5,5));
    collectionListPanel.setLayout(new FlowLayout(FlowLayout.LEFT));
    collectionLabel = new JLabel("Collection: ");

    collectionList = new JComboBox(csModel.getCollectionList()){
	       public boolean isFocusTraversable() {
		return false;
		}
    };
    collectionList.setEditable(false);
     
    collectionInfoButton = new JButton("Info...") {
	    public String getToolTipText(MouseEvent e) {
		return "Collection information for '" + csModel.getLongCollectionName((String)collectionList.getSelectedItem()) +"'";
	    }
    };
    ToolTipManager.sharedInstance().registerComponent(collectionInfoButton);
    collectionInfoButton.addActionListener(this);

    collectionListPanel.add(collectionLabel);
    collectionListPanel.add(collectionList);
    collectionListPanel.add(collectionInfoButton);

    searchTextFieldPanel = new JPanel();
    searchTextFieldPanel.setLayout(new FlowLayout(FlowLayout.LEFT));
    searchTextFieldPanel.add(searchTextField);

    //NzdlPreferences prefs = NzdlPreferences.getInstance();
    String queryType = NzdlConstants.DEFAULT_QUERY_TYPE;
    if (NzdlPreferences.getInstance().isString(NzdlConstants.QUERYTYPE)) {
	queryType = NzdlPreferences.getInstance().getString(NzdlConstants.QUERYTYPE);
    }
    // check integrity of queryType
    if (!((queryType.equals(NzdlConstants.RANKED_STRING)) | (queryType.equals(NzdlConstants.BOOLEAN_STRING)))) {
	// preferences may have been corrupted!
	queryType = NzdlConstants.DEFAULT_QUERY_TYPE;
    }
    boolean stemming = NzdlConstants.DEFAULT_STEMMING;
    if (NzdlPreferences.getInstance().isBoolean(NzdlConstants.STEMMING))
	stemming = NzdlPreferences.getInstance().getBoolean(NzdlConstants.STEMMING);
    boolean caseFolding = NzdlConstants.DEFAULT_CASE_FOLDING;
    if (NzdlPreferences.getInstance().isBoolean(NzdlConstants.CASE_FOLDING))
	caseFolding = NzdlPreferences.getInstance().getBoolean(NzdlConstants.CASE_FOLDING);
	
    optionsPanel = new JPanel();

    queryTypePanel = new JPanel();
    queryTypePanel.setLayout(new GridLayout(2,1));
    rankedRadioButton = new JRadioButton(NzdlConstants.RANKED_STRING);
    rankedRadioButton.setActionCommand(NzdlConstants.RANKED_STRING);
    rankedRadioButton.setSelected(queryType.equals(rankedRadioButton.getText()));
    rankedRadioButton.setToolTipText("Display results in a ranked list");
    
    booleanRadioButton = new JRadioButton(NzdlConstants.BOOLEAN_STRING);
    booleanRadioButton.setActionCommand(NzdlConstants.BOOLEAN_STRING);
    booleanRadioButton.setSelected(queryType.equals(booleanRadioButton.getText()));
    booleanRadioButton.setToolTipText("Allows the use of Boolean operators: AND(&) OR(|) and NOT(!)");
    buttonGroup = new ButtonGroup();
    buttonGroup.add(rankedRadioButton);
    buttonGroup.add(booleanRadioButton);
    queryTypePanel.add(rankedRadioButton);
    queryTypePanel.add(booleanRadioButton);

    stemCheckBox = new JCheckBox("Stemming", stemming);
    stemCheckBox.setToolTipText("Strip endings such as '...ing', '...ed'");
    caseFoldCheckBox = new JCheckBox("Match case", caseFolding);
    caseFoldCheckBox.setToolTipText("Only match when the case is the same");

    optionsPanel.add(queryTypePanel);
    optionsPanel.add(stemCheckBox);
    optionsPanel.add(caseFoldCheckBox);

    searchButtonPanel = new JPanel();
    searchButtonPanel.setLayout(new FlowLayout(FlowLayout.RIGHT));
    searchButtonPanel.add(searchButton);

    searchControlPanel = new JPanel();
    searchControlPanel.setLayout(new BorderLayout());
    searchControlPanel.add(optionsPanel, BorderLayout.WEST);
    searchControlPanel.add(searchButtonPanel, BorderLayout.EAST);

    queryFormulationPanel = new JPanel();
    queryFormulationPanel.setBorder(BorderFactory.createEmptyBorder(5,5,5,5));
    queryFormulationPanel.setLayout(new BoxLayout(queryFormulationPanel, BoxLayout.Y_AXIS));
    queryFormulationPanel.add(Box.createHorizontalStrut(100));
    queryFormulationPanel.add(collectionListPanel);
    queryFormulationPanel.add(searchTextFieldPanel);
    queryFormulationPanel.add(searchControlPanel);

    resultsPanel = new JPanel();
    resultsPanel.setLayout(new BorderLayout());
    resultsPanel.setBorder(BorderFactory.createEmptyBorder(5,5,5,5));
    resultListModel = csModel.getResultsModel();
    resultsList = new JList(resultListModel); 
    resultsList.setFont(resultsFont);
    resultsList.setBorder(BorderFactory.createEmptyBorder(5,5,5,5));
    resultsList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
    scrollResultsPane = new JScrollPane(resultsList);
    resultsPanel.add(scrollResultsPane, BorderLayout.CENTER);
    csModel.addResult(Result.INITIAL_FAKE_RESULT);

    resultPopup = new JPopupMenu();
    JMenuItem addBerryItem = new JMenuItem("Add to BerryBasket");
    addBerryItem.addActionListener(new ActionListener() {
	    public void actionPerformed(ActionEvent e) {
		System.err.println("SelectedResult = " + selectedResult);
		//System.err.println("add berry now...");
		csModel.getBerryBasketModel().add( new Berry(selectedResult.getQueryHistoryItem(),selectedResult.getDocID(),selectedResult.getCollectionName(), selectedResult.toString()));
	    } // actionPerformed
    });
    resultPopup.add(addBerryItem);
    MouseListener popupListener = new PopupListener();
    resultsList.addMouseListener(popupListener);




    dataPanel = new JPanel();
    dataPanel.setLayout(new BorderLayout());
    dataPanel.setBorder(BorderFactory.createEmptyBorder(3,3,3,3));
    documentPane = new JTextPane() { // to control scrolling for raw text docs
                           public boolean getScrollableTracksViewportWidth() { 
                           if (getSize().width < getParent().getSize().width)
			        return true; 
                           else return false; 
                           } 
                           public void setSize(Dimension d) { 
                           if (d.width < getParent().getSize().width) 
			        d.width = getParent().getSize().width; 
                                super.setSize(d); 
                           } 
	}; // end custom JTextPane

    documentPane.setEditable(false);
    documentPane.setFont(docFont);
    documentPane.setBorder(BorderFactory.createEmptyBorder(5,5,5,5));
    //documentPane.setPreferredSize(new Dimension(100, 100));

    scrollDataPane = new JScrollPane(documentPane);
    scrollDataPane.setPreferredSize(new Dimension(300, 300));
    dataPanel.add(scrollDataPane, BorderLayout.CENTER);

    parent.getRootPane().setDefaultButton(searchButton);

    resultsList.addListSelectionListener(new ResultListSelectionHandler() );

    add(Box.createHorizontalStrut(400));
    add(queryFormulationPanel);
    add(resultsPanel);
    add(dataPanel);
    searchTextField.requestFocus();
  }

  /** respond to the user pressing the Search button */

  public void actionPerformed(ActionEvent e) {
      String collectionName = collectionList.getSelectedItem().toString();
    if (e.getSource() == searchButton ) {
      // cursor not set when RETURN pressed, only when button clicked ?
      windowParent.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
      String queryString = searchTextField.getText();
      System.err.println("Search started (" + collectionName + "): \"" + queryString + "\" ..." );
      //send query to collection
      NzdlQuery nzdlQuery = new NzdlQuery(queryString);
      nzdlQuery.setQueryType(buttonGroup.getSelection().getActionCommand());
      //nzdlQuery.setQueryType("boolean");
      nzdlQuery.setStemming(stemCheckBox.isSelected());
      // interface reverses underlying logic so negate UI element state
      nzdlQuery.setCaseFolding(! caseFoldCheckBox.isSelected());

      // return the first numResults that match
      nzdlQuery.setEndResults( -1 );
      // "-1" means consider all the documents that match
      nzdlQuery.setMaxDocs( -1 );

      NzdlRequest request = new NzdlRequest( nzdlQuery );
      NzdlResponse response = new NzdlResponse();
      
      nzdl.service( collectionName, request, response );
      NzdlResultSet results = response.getResultSet();
      ArrayList docIDs = new ArrayList(results.getDocumentIDs());
      // System.err.println("Size of docIDs = " + docIDs.size());
      // paired collection of docIDs and Titles - but as Sets!
      Map titleMetaData = nzdl.getMetaData( collectionName, docIDs, "Title" );
      // created a paired collection of docIDs and Titles - as simple Strings!
      HashMap titleMap = new HashMap();
      for (ListIterator i = docIDs.listIterator(); i.hasNext(); ) {
	String docID = (String) i.next();
	//Set titleSet = (Set) titleMetaData.get( docID );
	ArrayList titleList = new ArrayList((Set)titleMetaData.get(docID));
	// use the first title as *the* title
	titleMap.put(docID, titleList.get(0)); 
      } // end for
      // update the results list
      csModel.clearResults();

      QueryHistoryItem queryHistoryItem =  new QueryHistoryItem(new Query(queryString), new Date(), collectionName, "user",docIDs.size(),buttonGroup.getSelection().getActionCommand(),stemCheckBox.isSelected(), caseFoldCheckBox.isSelected());
    csModel.getQueryHistoryModel().add(queryHistoryItem);

      if (docIDs.size() == 0 ) { // give some feedback for 0 results
	csModel.addResult(Result.FAKE_RESULT);
	documentPane.setStyledDocument(blankDoc);
      }
      else { // there are some results
	Iterator docIDsIterator = docIDs.iterator();
	while (docIDsIterator.hasNext()) {
	  String resultDocID =  (String) docIDsIterator.next();
	  csModel.addResult(new Result((String)titleMap.get(resultDocID), resultDocID, collectionName, queryHistoryItem ));
	} //end while
	if (NzdlPreferences.getInstance().getBoolean(NzdlConstants.DISPLAY_FIRST_DOC))
	resultsList.setSelectedIndex(0); 
      } // end else

      /*
      csModel.getQueryHistoryModel().add( new QueryHistoryItem(new Query(queryString), new Date(), collectionName, "user",docIDs.size(),buttonGroup.getSelection().getActionCommand(),stemCheckBox.isSelected(), caseFoldCheckBox.isSelected() ));
      */
      titleMap.clear();
      titleMetaData.clear();
      docIDs.clear();
      windowParent.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
    } // end if
    else {
	if (e.getSource() == collectionInfoButton) {
	    CollectionInfoDialog cid = new CollectionInfoDialog(windowParent, collectionName, csModel);
	}
	else {
	    System.err.println("unknown Action event in SearchPanel");
	}
    }
  } //end actionPerformed

  /** handles selection events in the results list so that when the 
      user selects a new title (say) the document contents is 
      automatically displayed in the main panel */

  class ResultListSelectionHandler implements ListSelectionListener {

      // private Result previousSelection = null;

    public void valueChanged(ListSelectionEvent e) {
      // get the result object from the SelectedIndex
      // assume we only get here with events from resultsList
	int docLength = 0;
      if (e.getValueIsAdjusting()) { return; }
      Result result =  (Result) resultsList.getSelectedValue();
      if ((result == Result.FAKE_RESULT) || (result == Result.INITIAL_FAKE_RESULT)) {
	  documentPane.setStyledDocument( blankDoc );
	  System.err.println("clicked on fake result");
      }
      else { // a real result representing a real document to deal with

	  windowParent.setCursor(WAIT_CURSOR);

	  deleteDocContents(); // for memory/garbage collection

	  StringReader sr = new StringReader(nzdl.getDocument(result.getCollectionName(),  result.getDocID()));

	  if (NzdlPreferences.getInstance().getBoolean(NzdlConstants.RAW_TEXT)) 
	      // display the document as raw text
	      displayAsRawText(sr);
	  else {
	      // treat as HTML
	      displayAsHTML(sr);
	      //htmlDoc.getImageData(nzdl, result.getCollectionName());
	  }

	  windowParent.setCursor(NORMAL_CURSOR);
      }
    } //end valueChanged
  }//end ResultListSelectionHandler



    private void deleteDocContents() {
	//trash the old htmlDoc for memory/garbage collection
	try {
	    if (htmlDoc != null)
		htmlDoc.remove(0, htmlDoc.getLength());
	    if (defaultStyledDoc != null)
		defaultStyledDoc.remove(0, defaultStyledDoc.getLength());
	}
	catch (Exception exception) {
	    throw new Error (exception.toString());
	}
    } // end deleteDocContents


    public void switchToRawText() {
	windowParent.setCursor(WAIT_CURSOR);
	System.err.println("switching to raw text");
	displayAsRawText(new StringReader(documentPane.getText()));
	windowParent.setCursor(NORMAL_CURSOR);
    }

    public void displayAsRawText(StringReader sr) {
	    defaultStyledDoc = new DefaultStyledDocument();
	    documentPane.setEditorKit(styledEditorKit);
	    try {
		styledEditorKit.read( sr, defaultStyledDoc, 0);
	    }
	    catch (Exception e) {
		System.err.println(e);
	    }
	    documentPane.setDocument(defaultStyledDoc);
	    setCaretToStart(defaultStyledDoc.getLength());
     } // end displayAsRawText


    public void switchToHTML() {
	     windowParent.setCursor(WAIT_CURSOR);
	     System.err.println("switching to HTML");
	     displayAsHTML(new StringReader(documentPane.getText()));
	     windowParent.setCursor(NORMAL_CURSOR);
    }

     public void displayAsHTML(StringReader sr) {
	     htmlDoc = new GMLDocument();
	     documentPane.setEditorKit(htmlEditorKit);
	     try {
		 htmlEditorKit.read( sr, htmlDoc, 0);
	     }
	     catch (Exception e) {
		 System.err.println(e);
	     }
	     documentPane.setStyledDocument( htmlDoc );
	     setCaretToStart(htmlDoc.getLength());
      } // end displayAsHTML


    private void setCaretToStart(int docLength) {
		 // it seems as if they haven't considered 
		 // documents with > MAX_INT num of chars ...
	if (docLength > 0)
		documentPane.setCaretPosition(1);
    }

    /*
     * save prefs not in preferencesDialog 
     */
    public void savePrefs() {
	NzdlPreferences.getInstance().setBoolean(NzdlConstants.STEMMING,stemCheckBox.isSelected());
	NzdlPreferences.getInstance().setBoolean(NzdlConstants.CASE_FOLDING,caseFoldCheckBox.isSelected());
	NzdlPreferences.getInstance().setString(NzdlConstants.QUERYTYPE,buttonGroup.getSelection().getActionCommand());
    } //savePrefs


    class PopupListener extends MouseAdapter {
	public void mousePressed(MouseEvent e) {
	    maybeShowPopup(e);
	}

	public void mouseReleased(MouseEvent e) {
	    maybeShowPopup(e);
	}

	private void maybeShowPopup(MouseEvent e) {
	    if (e.isPopupTrigger()) {
		int index = resultsList.locationToIndex(new Point(e.getX(), e.getY()));
		if (index < resultListModel.getSize()) {
		    Result result = (Result) resultListModel.getElementAt(index);
		    if (result != Result.INITIAL_FAKE_RESULT) {
			selectedResult = result;
			resultPopup.show(e.getComponent(),
					 e.getX(), e.getY());
		    }
		}
		else
		    System.err.println("popup not on item");
	    }
	}
    } // PopupListener



} //end SearchPanel
